from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .button import click_button
from .button_if_contains_text import click_button_if_contains_text


class ClickDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def button(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Click button."""

        click_button(self._driver, xpath, timeout)

    def button_if_contains_text(self, xpath: str, regex: str, ignore_case: bool = True, timeout: int = timeout.DEFAULT) -> None:

        click_button_if_contains_text(self._driver, xpath, regex, ignore_case, timeout)



class XPathSyntaxError(Exception):
    def __init__(self, xpath: str) -> None:
        self.message = f"Invalid XPath syntax: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from dataclasses import dataclass

from ... import helper
from ...model.type.xpath import XPath


@dataclass(kw_only=True, slots=True)
class SearchSettings:
    """Object with data needed to accept or decline cookies from a banner.

    input_xpath: XPath for the search input field.

    button_xpath: XPath for the search button.

    url: Optional URL for the search page.

    await_search_results_url_contains: Optional wait for the search results page URL to change.

    await_search_results_xpath: Optional wait for a search result element to be ready."""

    input_xpath: str
    button_xpath: str
    url: str | None = None
    await_search_results_url_contains: str | None = None
    await_search_results_xpath: str | None = None

    def __post_init__(self) -> None:
        self.input_xpath = XPath(self.input_xpath)
        self.button_xpath = XPath(self.button_xpath)
        self.url = helper.url.mediate_conversion_to_tiny_type_or_none(self.url)
        self.await_search_results_xpath = helper.xpath.mediate_conversion_to_tiny_type_or_none(
            self.await_search_results_xpath)



from ... import helper
from ...model.window.controller import WindowHandleController


def switch_to_window(driver: object, controller: WindowHandleController, window_handle: str) -> None:
    if helper.window_handle.is_valid_id(window_handle):
        driver.switch_to.window(window_handle)  # type: ignore
    else:
        window_handle_id = controller.get_handle_id_by_name(window_handle)
        driver.switch_to.window(window_handle_id)  # type: ignore



import re

from ... import constant, helper
from ...constant import timeout
from ...model.type.xpath import XPath
from ..get.text.from_element import get_text_from_element
from .for_element import wait_for_element


def wait_until_text_is(driver: object, xpath: str, regex: str, timeout: int = timeout.DEFAULT) -> None:
    def is_element_text(driver: object, xpath: str, regex: str) -> bool:
        text = get_text_from_element(driver, xpath, constant.timeout.BYPASS)
        return bool(re.match(regex, text))

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    helper.retry.until_condition_is_true(driver, xpath, regex, func=is_element_text, timeout=timeout)



import re

from ...constant import timeout
from ...model.type.xpath import XPath
from ..get.text.from_element import get_text_from_element


def check_if_contains_text(driver: object, xpath: str, regex: str, ignore_case: bool = True, timeout: int = timeout.DEFAULT) -> bool:
    xpath = XPath(xpath)
    current_text = get_text_from_element(driver, xpath, timeout)
    if ignore_case:
        match = re.search(regex, current_text, re.IGNORECASE)
    else:
        match = re.search(regex, current_text)
    return bool(match)



from ....constant import timeout
from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .from_element import get_text_from_element
from .from_elements import get_text_from_elements


class GetTextDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def from_element(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get text from element.

        This method assumes that the text field shouldn't be empty and therefore will retry to get the text (for better support of single-page apps with extended loading time)."""

        return get_text_from_element(self._driver, xpath, timeout)

    def from_elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[str]:
        """Get array of texts from elements.

        Assumes that the XPath targets multiple elements."""

        return get_text_from_elements(self._driver, xpath, timeout)



from .browser.base.driver import BrowserDriver
from .browser.base.settings import BrowserSettings


class DriverMethods:
    """Super class with initializer that can be extended in sub classes for web driver methods."""

    __slots__ = ["_browser_driver", "_driver", "_settings"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        self._browser_driver: BrowserDriver = browser_driver
        self._driver: object = browser_driver.webdriver
        self._settings: BrowserSettings = settings



from ..model.type.callable import DriverGetBoolCallable, DriverGetTextCallable


class RetryTimeoutException(Exception):
    def __init__(self, func: DriverGetBoolCallable | DriverGetTextCallable) -> None:
        self.message = f"Retry function timed out as loop ran out of retries for this function: {func}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from ... import helper
from ...model.browser.base.settings import BrowserSettings


def get_screenshot(driver: object, settings: BrowserSettings, file_name: str | None = None, destination_dir: str | None = None) -> None:
    if file_name is None:
        file_name = helper.screenshot.default_file_name()
    if destination_dir is None:
        destination_dir = settings.screenshot_dir
    else:
        destination_dir = helper.directory.ensure_trailing_slash(destination_dir)
        helper.directory.create_if_not_exists(destination_dir)
    driver.save_screenshot(f"{destination_dir}{file_name}")  # type: ignore



from selenium.webdriver.common.by import By

from .... import helper
from ....constant import timeout
from ....model.type.xpath import XPath
from ...wait.for_element import wait_for_element


def get_text_from_element(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> str:
    def get_inner_text_of_element(driver: object, xpath: str) -> str:
        return driver.find_element(By.XPATH, xpath).text  # type: ignore

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    return helper.retry.get_text_from_element(driver, xpath, get_inner_text_of_element)



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .url import open_url
from .url_if_not_current import open_url_if_not_current


class OpenDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def url(self, url: str) -> None:
        """Open web page by URL."""

        open_url(self._driver, url)

    def url_if_not_current(self, url: str, ignore_trailing_slash: bool = True, ignore_parameters: bool = False, ignore_https: bool = False) -> None:
        """Open a URL if it isn't already the current URL. Useful when doing multiple operations on a page where.

        ignore_trailing_slash: Ignore whether the URL is "http://example.com" or "http://example.com/".

        ignore_parameters: Ignore parameters in the URL, e.g. "http://example.com/list?page=1".

        ignore_https: Ignore whether the URL is "http://example.com" or "https://example.com"."""

        open_url_if_not_current(self._driver, url, ignore_trailing_slash, ignore_parameters, ignore_https)



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from ....model.window.controller import WindowHandleController
from .all import get_all_window_handles
from .count import count_window_handles
from .current import get_current_window_handle


class WindowHandleDriverMethods(DriverMethods):
    __slots__ = ["_controller"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings, controller: WindowHandleController) -> None:
        super().__init__(browser_driver, settings)
        self._controller = controller

    def all(self, selenium: bool = False) -> list[str]:
        """Get list of IDs of all open tabs or windows.

        selenium: Set to "True" if you want the Selenium handle IDs rather than from the internal window handle controller."""

        return get_all_window_handles(self._driver, self._controller, selenium)

    def count(self, selenium: bool = False) -> int:
        """Count number of open tabs or windows.

        selenium: Set to "True" if you want the number of Selenium handle IDs rather than from the internal window handle controller."""

        return count_window_handles(self._driver, self._controller, selenium)

    def current(self) -> str:
        """Get the ID of the current tab or window.

        Example: CDwindow-69663F4BF867CC38F6AF46D55BFC1A8A"""

        return get_current_window_handle(self._driver)



from datetime import datetime


def default_file_name() -> str:
    """Example: \"Browserist screenshot 2022-02-12 at 22.12.34.png\""""

    date = datetime.now().strftime("%Y-%m-%d")
    time = datetime.now().strftime("%H.%M.%S")
    return f"Browserist screenshot {date} at {time}.png"



from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.ie.options import Options as IEOptions
from selenium.webdriver.safari.options import Options as SafariOptions

from ..model.browser.base.driver import BrowserDriver


def page_load_strategy(browser_driver: BrowserDriver, options: ChromeOptions | EdgeOptions | FirefoxOptions | IEOptions | SafariOptions) -> ChromeOptions | EdgeOptions | FirefoxOptions | IEOptions | SafariOptions:
    options.page_load_strategy = browser_driver.settings.page_load_strategy.value
    return options



from ....constant import timeout
from ....model.type.xpath import XPath
from ..attribute.values import get_attribute_values


def get_url_from_multiple_images(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> list[str]:
    xpath = XPath(xpath)
    return get_attribute_values(driver, xpath, "src", timeout)



def check_if_is_image_element_loaded(driver: object, element: object) -> bool:
    is_image_loaded: bool = driver.execute_script(  # type: ignore
        "return arguments[0].complete && typeof arguments[0].naturalWidth != 'undefined' && arguments[0].naturalWidth > 0;",
        element)
    return is_image_loaded or False



from enum import Enum, unique


@unique
class TabOrWindow(Enum):
    """Selector to open new tab or window.

    TAB: Use for new tab.

    CONTINUE: Use for new window."""

    TAB = "tab"
    WINDOW = "window"



from .... import helper
from ....model.window.controller import WindowHandleController
from ....model.window.tab_or_window import TabOrWindow
from ...open.url import open_url
from ...wait.until_number_of_window_handles_is import wait_until_number_of_window_handles_is
from ..handle.count import count_window_handles
from ..handle.current import get_current_window_handle


def open_new_tab_or_window(driver: object,
                           controller: WindowHandleController,
                           tab_or_window: TabOrWindow,
                           url: str | None = None,
                           name: str | None = None) -> None:
    url = helper.url.mediate_conversion_to_tiny_type_or_none(url)
    current_number_of_window_handles = count_window_handles(driver, controller, selenium=True)
    driver.switch_to.new_window(tab_or_window.value)  # type: ignore
    wait_until_number_of_window_handles_is(driver, current_number_of_window_handles + 1)
    new_handle_id = get_current_window_handle(driver)
    controller.add_handle(new_handle_id, name)
    if url:
        open_url(driver, url)



from ...model.combo_settings.search import SearchSettings
from ..click.button import click_button
from ..input.value import input_value
from ..open.url_if_not_current import open_url_if_not_current
from ..wait.for_element import wait_for_element
from ..wait.until_url_contains import wait_until_url_contains


def combo_search(driver: object, term: str, settings: SearchSettings) -> None:
    if settings.url is not None:
        open_url_if_not_current(driver, settings.url)
    input_value(driver, settings.input_xpath, term)
    click_button(driver, settings.button_xpath)
    if settings.await_search_results_url_contains is not None:
        wait_until_url_contains(driver, settings.await_search_results_url_contains)
    if settings.await_search_results_xpath is not None:
        wait_for_element(driver, settings.await_search_results_xpath)



from ....constant import timeout
from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .current import get_current_url
from .from_image import get_url_from_image
from .from_link import get_url_from_link
from .from_multiple_images import get_url_from_multiple_images
from .from_multiple_links import get_url_from_multiple_links


class GetUrlDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def current(self) -> str:
        """Get URL of the current page."""

        return get_current_url(self._driver)

    def from_image(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get URL source from image, e.g. <img> tag.

        This method assumes that the image shouldn't be empty and therefore will retry to get the URL (for better support of single-page apps with extended loading time)."""

        return get_url_from_image(self._driver, xpath, timeout)

    def from_link(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get URL from link, e.g. <a> tag or button.

        This method assumes that the link shouldn't be empty and therefore will retry to get the URL (for better support of single-page apps with extended loading time)."""

        return get_url_from_link(self._driver, xpath, timeout)

    def from_multiple_images(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[str]:
        """Get array of URLs from images, e.g. <img> tags. Assumes that the XPath targets multiple images."""

        return get_url_from_multiple_images(self._driver, xpath, timeout)

    def from_multiple_links(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[str]:
        """Get array of URLs from links, e.g. <a> tags or buttons. Assumes that the XPath targets multiple links."""

        return get_url_from_multiple_links(self._driver, xpath, timeout)



from dataclasses import dataclass
from enum import Enum, auto

from ..constant import timeout


class TimeoutStrategy(Enum):
    """Define timeout strategy.

    BREAK: If something times out, stop operation.

    CONTINUE: If something times out, continue operation."""

    BREAK = auto()
    CONTINUE = auto()


@dataclass(slots=True)
class TimeoutSettings:
    strategy: TimeoutStrategy = TimeoutStrategy.BREAK
    seconds: int = timeout.DEFAULT
    _is_timed_out: bool = False



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .get_position import get_scroll_position
from .into_view import scroll_into_view
from .into_view_if_not_visible import scroll_into_view_if_not_visible
from .to_end_of_page import scroll_to_end_of_page
from .to_position import scroll_to_position
from .to_top_of_page import scroll_to_top_of_page


class ScrollDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def get_position(self) -> tuple[int, int]:
        """Get scroll position of the X and Y axis. Usage:

        x, y = browser.scroll.get_position()"""

        return get_scroll_position(self._driver)

    def into_view(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Find element and scroll up or down until element is visible."""

        scroll_into_view(self._driver, xpath, timeout)

    def into_view_if_not_visible(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """If not visilbe, find element and scroll up or down until element is visible."""

        scroll_into_view_if_not_visible(self._driver, xpath, timeout)

    def to_end_of_page(self) -> None:
        """Scroll to end of page."""

        scroll_to_end_of_page(self._driver)

    def to_position(self, x: int, y: int) -> None:
        """Scroll to coordinate X and Y pixels of page."""

        scroll_to_position(self._driver, x, y)

    def to_top_of_page(self) -> None:
        """Scroll to top of page."""

        scroll_to_top_of_page(self._driver)



import time

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By

from ... import constant
from ...constant import timeout
from ...exception.element import NoElementFoundException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def scroll_into_view(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    try:
        wait_for_element(driver, xpath, timeout)
        element: object = driver.find_element(By.XPATH, xpath)  # type: ignore
        driver.execute_script("arguments[0].scrollIntoView();", element)  # type: ignore
        time.sleep(constant.timeout.SHORT)  # Small timeout to make sure screen is updated after scroll.
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath) from NoSuchElementException



from .... import constant, helper
from ....constant import timeout
from ....model.type.xpath import XPath
from ...wait.for_element import wait_for_element
from ..attribute.value import get_attribute_value


def get_url_from_link(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> str:
    def get_href_attribute_of_element(driver: object, xpath: str) -> str:
        return get_attribute_value(driver, xpath, "href", constant.timeout.BYPASS)

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    return helper.retry.get_text_from_element(driver, xpath, get_href_attribute_of_element)



from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from ...model.type.xpath import XPath


def check_if_is_enabled(driver: object, xpath: str) -> bool:
    xpath = XPath(xpath)
    try:
        element = driver.find_element(By.XPATH, xpath)  # type: ignore
        return element.is_enabled()  # type: ignore
    except (NoSuchElementException, Exception):
        return False



from selenium.common.exceptions import NoSuchElementException, TimeoutException

from ...constant import timeout
from ...exception.element import NoElementFoundException, NoElementFoundWithTextConditionException
from ...exception.timeout import WaitForElementTimeoutException
from ...model.type.xpath import XPath
from ..check_if.contains_text import check_if_contains_text
from ..wait.for_element import wait_for_element
from .button import click_button


def click_button_if_contains_text(driver: object, xpath: str, regex: str, ignore_case: bool = True, timeout: int = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    try:
        wait_for_element(driver, xpath, timeout)
        if check_if_contains_text(driver, xpath, regex, ignore_case, timeout):
            click_button(driver, xpath, timeout)
        else:
            raise NoElementFoundWithTextConditionException(driver, xpath, regex)
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath) from TimeoutException
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath) from NoSuchElementException



from dataclasses import dataclass

from ... import helper
from ...model.type.xpath import XPath


@dataclass(kw_only=True, slots=True)
class CookieBannerSettings:
    """Object with data needed to accept or decline cookies from a banner.

    button_xpath: Can be for either accept or decline cookies.

    url: Optional URL from where to handle the cookie banner.

    has_loaded_wait_seconds: Minor grace time to make sure the cookie banner has loaded.

    has_loaded_xpath: Check if banner element has loaded so the cookie banner is ready.

    has_disappeared_wait_seconds: Minor grace time to make sure the cookie banner has disappeared and that the cookie information has been saved before proceeding."""

    button_xpath: str
    url: str | None = None
    has_loaded_wait_seconds: int | None = None
    has_loaded_xpath: str | None = None
    has_disappeared_wait_seconds: int | None = None

    def __post_init__(self) -> None:
        self.button_xpath = XPath(self.button_xpath)
        self.url = helper.url.mediate_conversion_to_tiny_type_or_none(self.url)
        self.has_loaded_xpath = helper.xpath.mediate_conversion_to_tiny_type_or_none(self.has_loaded_xpath)



from ..model.browser.base.driver import BrowserDriver


def disable_images(browser_driver: BrowserDriver) -> BrowserDriver:
    if browser_driver.settings.disable_images:
        preferences = {"profile.managed_default_content_settings.images": 2,
                       "profile.default_content_settings.images": 2}
        browser_driver.chrome_options.add_experimental_option("prefs", preferences)
    return browser_driver


def enable_headless(browser_driver: BrowserDriver) -> BrowserDriver:
    if browser_driver.settings.headless:
        browser_driver.chrome_options.headless = True
        browser_driver.chrome_options.add_argument("headless")  # type: ignore
    return browser_driver



from selenium import webdriver

from ... import factory
from .base.driver import BrowserDriver
from .base.type import BrowserType


class EdgeBrowserDriver(BrowserDriver):
    def ensure_browser_type(self) -> None:
        self.settings.type = BrowserType.EDGE

    def set_webdriver(self) -> object:
        if self.settings.path_to_executable is None:
            return webdriver.Edge(  # type: ignore
                options=self.edge_options)
        else:
            return webdriver.Edge(  # type: ignore
                executable_path=self.settings.path_to_executable,
                options=self.edge_options)

    def disable_images(self) -> None:
        if self.settings.disable_images:
            self.edge_options.use_chromium = True  # type: ignore
            preferences = {"profile.managed_default_content_settings.images": 2,
                           "profile.default_content_settings.images": 2}
            self.edge_options.add_experimental_option("prefs", preferences)

    def enable_headless(self) -> None:
        if self.settings.headless:
            self.edge_options.use_chromium = True  # type: ignore
            self.edge_options.add_argument("headless")  # type: ignore
            self.edge_options.add_argument("disable-gpu")  # type: ignore

    def set_page_load_strategy(self) -> None:
        self.edge_options = factory.set.page_load_strategy(self, self.edge_options)  # type: ignore



from selenium.webdriver.common.by import By

from ...constant import timeout
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element
from .is_image_element_loaded import check_if_is_image_element_loaded


def check_if_is_image_loaded(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> bool:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    element: object = driver.find_element(By.XPATH, xpath)  # type: ignore
    return check_if_is_image_element_loaded(driver, element)



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .clear import input_clear
from .value import input_value


class InputDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def clear(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Clear input form field."""

        input_clear(self._driver, xpath, timeout)

    def value(self, xpath: str, value: str, timeout: int = timeout.DEFAULT) -> None:
        """Input value into form field."""

        input_value(self._driver, xpath, value, timeout)



from typing import Any, Callable

DriverGetTextCallable = Callable[[object, str], str]

DriverGetBoolCallable = Callable[..., bool]

BrowserMethodWith2ArgumentsCallable = Callable[[object, str], None]

BrowserMethodWith3ArgumentsCallable = Callable[[object, str, Any], None]

BrowserMethodWith4ArgumentsCallable = Callable[[object, str, Any, Any], None]



from selenium import webdriver

from ... import factory
from ...exception.headless import HeadlessNotSupportedException
from .base.driver import BrowserDriver
from .base.type import BrowserType


class InternetExplorerBrowserDriver(BrowserDriver):
    def ensure_browser_type(self) -> None:
        self.settings.type = BrowserType.INTERNET_EXPLORER

    def set_webdriver(self) -> object:
        if self.settings.path_to_executable is None:
            return webdriver.Ie(  # type: ignore
                options=self.ie_options)
        else:
            return webdriver.Ie(  # type: ignore
                executable_path=self.settings.path_to_executable,
                options=self.ie_options)

    def disable_images(self) -> None:
        factory.internet_explorer.disable_images(self)

    def enable_headless(self) -> None:
        raise HeadlessNotSupportedException(self.settings.type)

    def set_page_load_strategy(self) -> None:
        self.ie_options = factory.set.page_load_strategy(self, self.ie_options)  # type: ignore



from ...constant import timeout
from ...model.type.xpath import XPath
from ..check_if.is_displayed import check_if_is_displayed
from .into_view import scroll_into_view


def scroll_into_view_if_not_visible(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    if not check_if_is_displayed(driver, xpath):
        scroll_into_view(driver, xpath, timeout)



import time

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from ...constant import timeout
from ...exception.element import NoElementFoundException


def scroll_to_end_of_page(driver: object) -> None:
    try:
        # Select the whole page before pressing any keys...
        body_element = driver.find_element(By.TAG_NAME, "body")  # type: ignore
        # ... and then simulate pressing End on the keyboard:
        body_element.send_keys(Keys.END)
        time.sleep(timeout.VERY_SHORT)  # Small delay to ensure the view is updated.
    except NoSuchElementException:
        raise NoElementFoundException(driver, "body") from NoSuchElementException



from ... import helper
from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .count_elements import tool_count_elements
from .is_input_valid import tool_is_input_valid


class ToolDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def is_input_valid(self, text: str, regex: str, ignore_case: bool = True) -> bool:
        """Check if input matches regex condition."""

        return tool_is_input_valid(text, regex, ignore_case)

    def is_url_valid(self, url: str) -> bool:
        """Check if input is a valid URL."""

        return helper.url.is_valid(url)

    def count_elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> int:
        """Count number of elements."""

        return tool_count_elements(self._driver, xpath, timeout)



from __future__ import annotations

from types import TracebackType

from .. import factory
from ..model.browser.base.driver import BrowserDriver
from ..model.browser.base.settings import BrowserSettings
from ..model.browser.base.type import BrowserType
from ..model.browser.extension.internet_explorer import InternetExplorerBrowserExtension
from ..model.browser.extension.safari import SafariBrowserExtension
from .check_if.__main__ import CheckIfDriverMethods
from .click.__main__ import ClickDriverMethods
from .combo.__main__ import ComboDriverMethods
from .get.__main__ import GetDriverMethods
from .iframe.__main__ import IframeDriverMethods
from .input.__main__ import InputDriverMethods
from .mouse.__main__ import MouseDriverMethods
from .open.__main__ import OpenDriverMethods
from .scroll.__main__ import ScrollDriverMethods
from .select.__main__ import SelectDriverMethods
from .tool.__main__ import ToolDriverMethods
from .wait.__main__ import WaitDriverMethods
from .window.__main__ import WindowDriverMethods


class Browser:
    """Main class of Browserist that sets the Selenium web driver and contains all helper functions."""

    __slots__ = ["_browser_driver", "driver", "ie", "safari",
                 "check_if", "click", "combo", "get", "iframe", "input", "mouse", "open", "scroll", "select", "tool", "wait", "window"]

    def __init__(self, settings: BrowserSettings | None = None) -> None:
        """Initiates the browser driver whether the settings calls for Chrome, Firefox, etc."""

        if settings is None:
            settings = BrowserSettings()  # Use default settings if no custom settings are given.

        self._browser_driver: BrowserDriver = factory.get.browser_driver(settings)
        self.driver: object = self._browser_driver.webdriver

        match self._browser_driver.settings.type:
            case BrowserType.INTERNET_EXPLORER:
                self.ie: InternetExplorerBrowserExtension = InternetExplorerBrowserExtension(self._browser_driver)
            case BrowserType.SAFARI:
                self.safari: SafariBrowserExtension = SafariBrowserExtension(self._browser_driver)
            case _:
                pass

        self.check_if: CheckIfDriverMethods = CheckIfDriverMethods(self._browser_driver, settings)
        self.click: ClickDriverMethods = ClickDriverMethods(self._browser_driver, settings)
        self.combo: ComboDriverMethods = ComboDriverMethods(self._browser_driver, settings)
        self.get: GetDriverMethods = GetDriverMethods(self._browser_driver, settings)
        self.iframe: IframeDriverMethods = IframeDriverMethods(self._browser_driver, settings)
        self.input: InputDriverMethods = InputDriverMethods(self._browser_driver, settings)
        self.mouse: MouseDriverMethods = MouseDriverMethods(self._browser_driver, settings)
        self.open: OpenDriverMethods = OpenDriverMethods(self._browser_driver, settings)
        self.scroll: ScrollDriverMethods = ScrollDriverMethods(self._browser_driver, settings)
        self.select: SelectDriverMethods = SelectDriverMethods(self._browser_driver, settings)
        self.tool: ToolDriverMethods = ToolDriverMethods(self._browser_driver, settings)
        self.wait: WaitDriverMethods = WaitDriverMethods(self._browser_driver, settings)
        self.window: WindowDriverMethods = WindowDriverMethods(self._browser_driver, settings)

    def __enter__(self) -> Browser:
        return self

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> None:
        self.quit()

    def back(self) -> None:
        """Press the browser's back button."""

        self.driver.back()  # type: ignore

    def forward(self) -> None:
        """Press the browser's forward button."""

        self.driver.forward()  # type: ignore

    def refresh(self) -> None:
        """Refresh the current page."""

        self.driver.refresh()  # type: ignore

    def quit(self) -> None:
        """Quit the browser."""

        self.driver.quit()  # type: ignore



from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from ...constant import timeout
from ...exception.timeout import WaitForPageTitleToChangeTimeoutException


def wait_until_page_title_contains(driver: object, page_title_fragment: str, timeout: int = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.title_contains(page_title_fragment))  # type: ignore
    except TimeoutException:
        raise WaitForPageTitleToChangeTimeoutException(driver, page_title_fragment) from TimeoutException



from ...model.type.xpath import XPath
from .is_enabled import check_if_is_enabled


def check_if_is_disabled(driver: object, xpath: str) -> bool:
    xpath = XPath(xpath)
    return not check_if_is_enabled(driver, xpath)



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .attribute.__main__ import GetAttributeDriverMethods
from .dimensions import get_dimensions
from .element import get_element
from .elements import get_elements
from .elements_by_tag import get_elements_by_tag
from .page_title import get_page_title
from .screenshot import get_screenshot
from .text.__main__ import GetTextDriverMethods
from .url.__main__ import GetUrlDriverMethods


class GetDriverMethods(DriverMethods):
    __slots__ = ["attribute", "text", "url"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self.attribute: GetAttributeDriverMethods = GetAttributeDriverMethods(browser_driver, settings)
        self.text: GetTextDriverMethods = GetTextDriverMethods(browser_driver, settings)
        self.url: GetUrlDriverMethods = GetUrlDriverMethods(browser_driver, settings)

    def dimensions(self, xpath: str, timeout: int = timeout.DEFAULT) -> tuple[int, int]:
        """Get width and height of element in pixels. Usage:

        width, height = browser.get.dimensions("/element/xpath")"""

        return get_dimensions(self._driver, xpath, timeout)

    def element(self, xpath: str, timeout: int = timeout.DEFAULT) -> object:
        """Get single web element by XPath."""

        return get_element(self._driver, xpath, timeout)

    def elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[object]:
        """Get multiple web elements by XPath."""

        return get_elements(self._driver, xpath, timeout)

    def elements_by_tag(self, tag: str, timeout: int = timeout.DEFAULT) -> list[object]:
        """"Get multiple web elements by HTML tag. Examples: "img" as tag for all <img> images, "a" for all <a> links, etc."""

        return get_elements_by_tag(self._driver, tag, timeout)

    def page_title(self) -> str:
        """Get page title of the current page."""

        return get_page_title(self._driver)

    def screenshot(self, file_name: str | None = None, destination_dir: str | None = None) -> None:
        """Take screenshot and save as PNG image. Default destination directory is from where the script is executed. Examples:

        browser.get.screenshot() # Default file name and destination

        browser.get.screenshot("image.png") # Custom file name and default destination

        browser.get.screenshot("image.png", "./screenshots") # Custom file name and destination

        browser.get.screenshot(destination_dir = "./screenshots") # Default file name and custom destination"""

        get_screenshot(self._driver, self._settings, file_name, destination_dir)



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .for_element import wait_for_element
from .random_time import wait_random_time
from .until_element_disappears import wait_until_element_disappears
from .until_images_have_loaded import wait_until_images_have_loaded
from .until_number_of_window_handles_is import wait_until_number_of_window_handles_is
from .until_page_title_contains import wait_until_page_title_contains
from .until_page_title_is import wait_until_page_title_is
from .until_text_changes import wait_until_text_changes
from .until_text_contains import wait_until_text_contains
from .until_text_is import wait_until_text_is
from .until_url_changes import wait_until_url_changes
from .until_url_contains import wait_until_url_contains
from .until_url_is import wait_until_url_is


class WaitDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def for_element(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element is ready in the DOM and/or on the screen.

        Especially useful for single-page app elements handled/modified by JavaScript, but also standard HTML that doesn't load immediately, this helper function ensures that DOM elements are ready before processing."""

        wait_for_element(self._driver, xpath, timeout)

    def random_time(self, min_seconds: int = 1, max_seconds: int = 5) -> None:
        """Randomize sleep timing to make actions look less like a bot."""

        wait_random_time(min_seconds, max_seconds)

    def until_element_disappears(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist."""

        wait_until_element_disappears(self._driver, xpath, timeout)

    def until_images_have_loaded(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist. The image XPath can target one or more images."""

        wait_until_images_have_loaded(self._driver, xpath, timeout)

    def until_number_of_window_handles_is(self, timeout: int = timeout.DEFAULT) -> None:
        """Wait until number of window handles is."""

        wait_until_number_of_window_handles_is(self._driver, timeout)

    def until_page_title_contains(self, page_title_fragment: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input can contain both a fragment or the full page title."""

        wait_until_page_title_contains(self._driver, page_title_fragment, timeout)

    def until_page_title_is(self, page_title: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input has to match the exact page title."""

        wait_until_page_title_is(self._driver, page_title, timeout)

    def until_text_changes(self, xpath: str, baseline_text: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the text of an element changes from a baseline text, e.g. after a form action. The text is evaluated as an exact match."""

        wait_until_text_changes(self._driver, xpath, baseline_text, timeout)

    def until_text_contains(self, xpath: str, regex: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the text of an element has changed, e.g. after a form action."""

        wait_until_text_contains(self._driver, xpath, regex, timeout)

    def until_text_is(self, xpath: str, regex: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the text of an element has changed, e.g. after a form action. The text is evaluated as an exact match."""

        wait_until_text_is(self._driver, xpath, regex, timeout)

    def until_url_changes(self, baseline_url: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed from a baseline URL, e.g. after a redirect or form action. The URL is evaluated as an exact match."""

        wait_until_url_changes(self._driver, baseline_url, timeout)

    def until_url_contains(self, url_fragment: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed, e.g. after a redirect. The URL variable can contain both a fragment (e.g. ?login=true) or a full URL (e.g. https://www.example.com/?login=true)."""

        wait_until_url_contains(self._driver, url_fragment, timeout)

    def until_url_is(self, url: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the browser URL has changed, e.g. after a redirect. The URL is evaluated as an exact match."""

        wait_until_url_is(self._driver, url, timeout)
